_begin_form = "begin_form"
_begin_file = "begin_file"
_cont = "cont"
_end = "end"


class MultiPartParser(object):
    def __init__(
        self,
        stream_factory=None,
        charset="utf-8",
        errors="replace",
        max_form_memory_size=None,
        cls=None,
        buffer_size=64 * 1024,
    ):
        self.charset = charset
        self.errors = errors
        self.max_form_memory_size = max_form_memory_size
        self.stream_factory = (
            default_stream_factory if stream_factory is None else stream_factory
        )
        self.cls = MultiDict if cls is None else cls

        # make sure the buffer size is divisible by four so that we can base64
        # decode chunk by chunk
        assert buffer_size % 4 == 0, "buffer size has to be divisible by 4"
        # also the buffer size has to be at least 1024 bytes long or long headers
        # will freak out the system
        assert buffer_size >= 1024, "buffer size has to be at least 1KB"

        self.buffer_size = buffer_size

    def _fix_ie_filename(self, filename):
        """Internet Explorer 6 transmits the full file name if a file is
        uploaded.  This function strips the full path if it thinks the
        filename is Windows-like absolute.
        """
        if filename[1:3] == ":\\" or filename[:2] == "\\\\":
            return filename.split("\\")[-1]
        return filename

    def _find_terminator(self, iterator):
        """The terminator might have some additional newlines before it.
        There is at least one application that sends additional newlines
        before headers (the python setuptools package).
        """
        for line in iterator:
            if not line:
                break
            line = line.strip()
            if line:
                return line
        return b""

    def fail(self, message):
        raise ValueError(message)

    def get_part_encoding(self, headers):
        transfer_encoding = headers.get("content-transfer-encoding")
        if (
            transfer_encoding is not None
            and transfer_encoding in _supported_multipart_encodings
        ):
            return transfer_encoding

    def get_part_charset(self, headers):
        # Figure out input charset for current part
        content_type = headers.get("content-type")
        if content_type:
            mimetype, ct_params = parse_options_header(content_type)
            return ct_params.get("charset", self.charset)
        return self.charset

    def start_file_streaming(self, filename, headers, total_content_length):
        if isinstance(filename, bytes):
            filename = filename.decode(self.charset, self.errors)
        filename = self._fix_ie_filename(filename)
        content_type = headers.get("content-type")
        try:
            content_length = int(headers["content-length"])
        except (KeyError, ValueError):
            content_length = 0
        container = self.stream_factory(
            total_content_length, content_type, filename, content_length
        )
        return filename, container

    def in_memory_threshold_reached(self, bytes):
        raise exceptions.RequestEntityTooLarge()

    def validate_boundary(self, boundary):
        if not boundary:
            self.fail("Missing boundary")
        if not is_valid_multipart_boundary(boundary):
            self.fail("Invalid boundary: %s" % boundary)
        if len(boundary) > self.buffer_size:  # pragma: no cover
            # this should never happen because we check for a minimum size
            # of 1024 and boundaries may not be longer than 200.  The only
            # situation when this happens is for non debug builds where
            # the assert is skipped.
            self.fail("Boundary longer than buffer size")

    class LineSplitter(object):
        def __init__(self, cap=None):
            self.buffer = b""
            self.cap = cap

        def _splitlines(self, pre, post):
            buf = pre + post
            rv = []
            if not buf:
                return rv, b""
            lines = buf.splitlines(True)
            iv = b""
            for line in lines:
                iv += line
                while self.cap and len(iv) >= self.cap:
                    rv.append(iv[: self.cap])
                    iv = iv[self.cap :]
                if line[-1:] in b"\r\n":
                    rv.append(iv)
                    iv = b""
            # If this isn't the very end of the stream and what we got ends
            # with \r, we need to hold on to it in case an \n comes next
            if post and rv and not iv and rv[-1][-1:] == b"\r":
                iv = rv[-1]
                del rv[-1]
            return rv, iv

        def feed(self, data):
            lines, self.buffer = self._splitlines(self.buffer, data)
            if not data:
                lines += [self.buffer]
                if self.buffer:
                    lines += [b""]
            return lines

    class LineParser(object):
        def __init__(self, parent, boundary):
            self.parent = parent
            self.boundary = boundary
            self._next_part = b"--" + boundary
            self._last_part = self._next_part + b"--"
            self._state = self._state_pre_term
            self._output = []
            self._headers = []
            self._tail = b""
            self._codec = None

        def _start_content(self):
            disposition = self._headers.get("content-disposition")
            if disposition is None:
                raise ValueError("Missing Content-Disposition header")
            self.disposition, extra = parse_options_header(disposition)
            transfer_encoding = self.parent.get_part_encoding(self._headers)
            if transfer_encoding is not None:
                if transfer_encoding == "base64":
                    transfer_encoding = "base64_codec"
                try:
                    self._codec = codecs.lookup(transfer_encoding)
                except Exception:
                    raise ValueError(
                        "Cannot decode transfer-encoding: %r" % transfer_encoding
                    )
            self.name = extra.get("name")
            self.filename = extra.get("filename")
            if self.filename is not None:
                self._output.append(
                    ("begin_file", (self._headers, self.name, self.filename))
                )
            else:
                self._output.append(("begin_form", (self._headers, self.name)))
            return self._state_output

        def _state_done(self, line):
            return self._state_done

        def _state_output(self, line):
            if not line:
                raise ValueError("Unexpected end of file")
            sline = line.rstrip()
            if sline == self._last_part:
                self._tail = b""
                self._output.append(("end", None))
                return self._state_done
            elif sline == self._next_part:
                self._tail = b""
                self._output.append(("end", None))
                self._headers = []
                return self._state_headers

            if self._codec:
                try:
                    line, _ = self._codec.decode(line)
                except Exception:
                    raise ValueError("Could not decode transfer-encoded chunk")

            # We don't know yet whether we can output the final newline, so
            # we'll save it in self._tail and output it next time.
            tail = self._tail
            if line[-2:] == b"\r\n":
                self._output.append(("cont", tail + line[:-2]))
                self._tail = line[-2:]
            elif line[-1:] in b"\r\n":
                self._output.append(("cont", tail + line[:-1]))
                self._tail = line[-1:]
            else:
                self._output.append(("cont", tail + line))
                self._tail = b""
            return self._state_output

        def _state_pre_term(self, line):
            if not line:
                raise ValueError("Unexpected end of file")
                return self._state_pre_term
            line = line.rstrip(b"\r\n")
            if not line:
                return self._state_pre_term
            if line == self._last_part:
                return self._state_done
            elif line == self._next_part:
                self._headers = []
                return self._state_headers
            raise ValueError("Expected boundary at start of multipart data")

        def _state_headers(self, line):
            if line is None:
                raise ValueError("Unexpected end of file during headers")
            line = to_native(line)
            line, line_terminated = _line_parse(line)
            if not line_terminated:
                raise ValueError("Unexpected end of line in multipart header")
            if not line:
                self._headers = Headers(self._headers)
                return self._start_content()
            if line[0] in " \t" and self._headers:
                key, value = self._headers[-1]
                self._headers[-1] = (key, value + "\n " + line[1:])
            else:
                parts = line.split(":", 1)
                if len(parts) == 2:
                    self._headers.append((parts[0].strip(), parts[1].strip()))
                else:
                    raise ValueError("Malformed header")
            return self._state_headers

        def feed(self, lines):
            self._output = []
            s = self._state
            for line in lines:
                s = s(line)
            self._state = s
            return self._output

    class PartParser(object):
        def __init__(self, parent, content_length):
            self.parent = parent
            self.content_length = content_length
            self._write = None
            self._in_memory = 0
            self._guard_memory = False

        def _feed_one(self, event):
            ev, data = event
            p = self.parent
            if ev == "begin_file":
                self._headers, self._name, filename = data
                self._filename, self._container = p.start_file_streaming(
                    filename, self._headers, self.content_length
                )
                self._write = self._container.write
                self._is_file = True
                self._guard_memory = False
            elif ev == "begin_form":
                self._headers, self._name = data
                self._container = []
                self._write = self._container.append
                self._is_file = False
                self._guard_memory = p.max_form_memory_size is not None
            elif ev == "cont":
                self._write(data)
                if self._guard_memory:
                    self._in_memory += len(data)
                    if self._in_memory > p.max_form_memory_size:
                        p.in_memory_threshold_reached(self._in_memory)
            elif ev == "end":
                if self._is_file:
                    self._container.seek(0)
                    return (
                        "file",
                        (
                            self._name,
                            FileStorage(
                                self._container,
                                self._filename,
                                self._name,
                                headers=self._headers,
                            ),
                        ),
                    )
                else:
                    part_charset = p.get_part_charset(self._headers)
                    return (
                        "form",
                        (
                            self._name,
                            b"".join(self._container).decode(part_charset, p.errors),
                        ),
                    )

        def feed(self, events):
            rv = []
            for event in events:
                v = self._feed_one(event)
                if v is not None:
                    rv.append(v)
            return rv

    def parse_lines(self, file, boundary, content_length, cap_at_buffer=True):
        """Generate parts of
        ``('begin_form', (headers, name))``
        ``('begin_file', (headers, name, filename))``
        ``('cont', bytestring)``
        ``('end', None)``
        Always obeys the grammar
        parts = ( begin_form cont* end |
                  begin_file cont* end )*
        """

        line_splitter = self.LineSplitter(self.buffer_size if cap_at_buffer else None)
        line_parser = self.LineParser(self, boundary)
        while True:
            buf = file.read(self.buffer_size)
            lines = line_splitter.feed(buf)
            parts = line_parser.feed(lines)
            for part in parts:
                yield part
            if buf == b"":
                break

    def parse_parts(self, file, boundary, content_length):
        """Generate ``('file', (name, val))`` and
        ``('form', (name, val))`` parts.
        """
        line_splitter = self.LineSplitter()
        line_parser = self.LineParser(self, boundary)
        part_parser = self.PartParser(self, content_length)
        while True:
            buf = file.read(self.buffer_size)
            lines = line_splitter.feed(buf)
            parts = line_parser.feed(lines)
            events = part_parser.feed(parts)
            for event in events:
                yield event
            if buf == b"":
                break

    def parse(self, file, boundary, content_length):
        formstream, filestream = tee(
            self.parse_parts(file, boundary, content_length), 2
        )
        form = (p[1] for p in formstream if p[0] == "form")
        files = (p[1] for p in filestream if p[0] == "file")
        return self.cls(form), self.cls(files)



from starlette import Response
from starlette.types import Scope, ASGIApp, ASGIInstance
import re
import typing


class Route:
    def matches(self, scope: Scope) -> typing.Tuple[bool, Scope]:
        raise NotImplementedError()  # pragma: no cover

    def __call__(self, scope: Scope) -> ASGIInstance:
        raise NotImplementedError()  # pragma: no cover


class Path(Route):
    def __init__(
        self, path: str, app: ASGIApp, methods: typing.Sequence[str] = ()
    ) -> None:
        self.path = path
        self.app = app
        self.methods = methods
        regex = "^" + path + "$"
        regex = re.sub("{([a-zA-Z_][a-zA-Z0-9_]*)}", r"(?P<\1>[^/]+)", regex)
        self.path_regex = re.compile(regex)

    def matches(self, scope: Scope) -> typing.Tuple[bool, Scope]:
        match = self.path_regex.match(scope["path"])
        if match:
            kwargs = dict(scope.get("kwargs", {}))
            kwargs.update(match.groupdict())
            child_scope = dict(scope)
            child_scope["kwargs"] = kwargs
            return True, child_scope
        return False, {}

    def __call__(self, scope: Scope) -> ASGIInstance:
        if self.methods and scope["method"] not in self.methods:
            return Response("Method not allowed", 406, media_type="text/plain")
        return self.app(scope)


class PathPrefix(Route):
    def __init__(
        self, path: str, app: ASGIApp, methods: typing.Sequence[str] = ()
    ) -> None:
        self.path = path
        self.app = app
        self.methods = methods
        regex = "^" + path
        regex = re.sub("{([a-zA-Z_][a-zA-Z0-9_]*)}", r"(?P<\1>[^/]*)", regex)
        self.path_regex = re.compile(regex)

    def matches(self, scope: Scope) -> typing.Tuple[bool, Scope]:
        match = self.path_regex.match(scope["path"])
        if match:
            kwargs = dict(scope.get("kwargs", {}))
            kwargs.update(match.groupdict())
            child_scope = dict(scope)
            child_scope["kwargs"] = kwargs
            child_scope["root_path"] = scope.get("root_path", "") + match.string
            child_scope["path"] = scope["path"][match.span()[1] :]
            return True, child_scope
        return False, {}

    def __call__(self, scope: Scope) -> ASGIInstance:
        if self.methods and scope["method"] not in self.methods:
            return Response("Method not allowed", 406, media_type="text/plain")
        return self.app(scope)


class Router:
    def __init__(self, routes: typing.List[Route], default: ASGIApp = None) -> None:
        self.routes = routes
        self.default = self.not_found if default is None else default

    def __call__(self, scope: Scope) -> ASGIInstance:
        for route in self.routes:
            matched, child_scope = route.matches(scope)
            if matched:
                return route(child_scope)
        return self.not_found(scope)

    def not_found(self, scope: Scope) -> ASGIInstance:
        return Response("Not found", 404, media_type="text/plain")



from functools import partial
from urllib.parse import urlparse
import asyncio
import asyncpg


class DatabaseMiddleware:
    def __init__(self, app, database_url=None, database_config=None):
        self.app = app
        if database_config is None:
            parsed = urlparse(database_url)
            database_config = {
                "user": parsed.user,
                "password": parsed.password,
                "database": parsed.database,
                "host": parsed.host,
                "port": parsed.port,
            }
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.create_pool(database_config))

    async def create_pool(self, database_config):
        self.pool = await asyncpg.create_pool(**database_config)

    def __call__(self, scope):
        scope["database"] = self.pool
        return self.app(scope)



from starlette.datastructures import URL, Headers, QueryParams
from collections.abc import Mapping
import json
import typing


class Request(Mapping):
    def __init__(self, scope, receive=None):
        self._scope = scope
        self._receive = receive
        self._stream_consumed = False

    def __getitem__(self, key):
        return self._scope[key]

    def __iter__(self):
        return iter(self._scope)

    def __len__(self):
        return len(self._scope)

    @property
    def method(self) -> str:
        return self._scope["method"]

    @property
    def url(self) -> URL:
        if not hasattr(self, "_url"):
            scheme = self._scope["scheme"]
            host, port = self._scope["server"]
            path = self._scope.get("root_path", "") + self._scope["path"]
            query_string = self._scope["query_string"]

            if (scheme == "http" and port != 80) or (scheme == "https" and port != 443):
                url = "%s://%s:%s%s" % (scheme, host, port, path)
            else:
                url = "%s://%s%s" % (scheme, host, path)

            if query_string:
                url += "?" + query_string.decode()

            self._url = URL(url)
        return self._url

    @property
    def headers(self) -> Headers:
        if not hasattr(self, "_headers"):
            self._headers = Headers(self._scope["headers"])
        return self._headers

    @property
    def query_params(self) -> QueryParams:
        if not hasattr(self, "_query_params"):
            query_string = self._scope["query_string"].decode()
            self._query_params = QueryParams(query_string)
        return self._query_params

    async def stream(self):
        if hasattr(self, "_body"):
            yield self._body
            return

        if self._stream_consumed:
            raise RuntimeError("Stream consumed")

        if self._receive is None:
            raise RuntimeError("Receive channel has not been made available")

        self._stream_consumed = True
        while True:
            message = await self._receive()
            if message["type"] == "http.request":
                yield message.get("body", b"")
                if not message.get("more_body", False):
                    break

    async def body(self):
        if not hasattr(self, "_body"):
            body = b""
            async for chunk in self.stream():
                body += chunk
            self._body = body
        return self._body

    async def json(self):
        if not hasattr(self, "_json"):
            body = await self.body()
            self._json = json.loads(body)
        return self._json



import asyncio
from starlette.request import Request
from starlette.response import Response
from starlette.types import ASGIInstance, Receive, Send, Scope


def asgi_application(func):
    is_coroutine = asyncio.iscoroutinefunction(func)

    def app(scope: Scope) -> ASGIInstance:
        async def awaitable(receive: Receive, send: Send) -> None:
            request = Request(scope, receive)
            if is_coroutine:
                response = await func(request)
            else:
                response = func(request)
            await response(receive, send)

        return awaitable

    return app



import typing
from urllib.parse import parse_qsl, urlparse


class URL(str):
    @property
    def components(self):
        if not hasattr(self, "_components"):
            self._components = urlparse(self)
        return self._components

    @property
    def scheme(self):
        return self.components.scheme

    @property
    def netloc(self):
        return self.components.netloc

    @property
    def path(self):
        return self.components.path

    @property
    def params(self):
        return self.components.params

    @property
    def query(self):
        return self.components.query

    @property
    def fragment(self):
        return self.components.fragment

    @property
    def username(self):
        return self.components.username

    @property
    def password(self):
        return self.components.password

    @property
    def hostname(self):
        return self.components.hostname

    @property
    def port(self):
        return self.components.port


# Type annotations for valid `__init__` values to QueryParams and Headers.
StrPairs = typing.Sequence[typing.Tuple[str, str]]
StrDict = typing.Mapping[str, str]


class QueryParams(typing.Mapping[str, str]):
    """
    An immutable multidict.
    """

    def __init__(
        self, value: typing.Union[str, typing.Union[StrDict, StrPairs]] = None
    ) -> None:
        if value is None:
            value = []
        elif isinstance(value, str):
            value = parse_qsl(value)

        if hasattr(value, "items"):
            items = list(typing.cast(StrDict, value).items())
        else:
            items = list(typing.cast(StrPairs, value))
        self._dict = {k: v for k, v in reversed(items)}
        self._list = items

    def get_list(self, key: str) -> typing.List[str]:
        return [item_value for item_key, item_value in self._list if item_key == key]

    def keys(self):
        return [key for key, value in self._list]

    def values(self):
        return [value for key, value in self._list]

    def items(self):
        return list(self._list)

    def get(self, key, default=None):
        if key in self._dict:
            return self._dict[key]
        else:
            return default

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def __iter__(self):
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def __eq__(self, other):
        if not isinstance(other, QueryParams):
            other = QueryParams(other)
        return sorted(self._list) == sorted(other._list)

    def __repr__(self):
        return "QueryParams(%s)" % repr(self._list)


class Headers(typing.Mapping[str, str]):
    """
    An immutable, case-insensitive multidict.
    """

    def __init__(self, value: typing.Union[StrDict, StrPairs] = None) -> None:
        if value is None:
            self._list = []
        else:
            assert isinstance(value, list)
            for header_key, header_value in value:
                assert isinstance(header_key, bytes)
                assert isinstance(header_value, bytes)
                assert header_key == header_key.lower()
            self._list = value

    def keys(self):
        return [key.decode("latin-1") for key, value in self._list]

    def values(self):
        return [value.decode("latin-1") for key, value in self._list]

    def items(self):
        return [
            (key.decode("latin-1"), value.decode("latin-1"))
            for key, value in self._list
        ]

    def get(self, key: str, default: str = None):
        try:
            return self[key]
        except KeyError:
            return default

    def get_list(self, key: str) -> typing.List[str]:
        get_header_key = key.lower().encode("latin-1")
        return [
            item_value.decode("latin-1")
            for item_key, item_value in self._list
            if item_key == get_header_key
        ]

    def __getitem__(self, key: str):
        get_header_key = key.lower().encode("latin-1")
        for header_key, header_value in self._list:
            if header_key == get_header_key:
                return header_value.decode("latin-1")
        raise KeyError(key)

    def __contains__(self, key: str):
        return key.lower() in self.keys()

    def __iter__(self):
        return iter(self.items())

    def __len__(self):
        return len(self._list)

    def __eq__(self, other):
        if not isinstance(other, Headers):
            return False
        return sorted(self._list) == sorted(other._list)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self.items()))


class MutableHeaders(Headers):
    def __setitem__(self, key: str, value: str):
        set_key = key.lower().encode("latin-1")
        set_value = value.encode("latin-1")

        pop_indexes = []
        for idx, (item_key, item_value) in enumerate(self._list):
            if item_key == set_key:
                pop_indexes.append(idx)

        for idx in reversed(pop_indexes):
            del self._list[idx]

        self._list.append((set_key, set_value))



from starlette.decorators import asgi_application
from starlette.response import (
    HTMLResponse,
    JSONResponse,
    Response,
    PlainTextResponse,
    StreamingResponse,
)
from starlette.request import Request
from starlette.routing import Path, PathPrefix, Router
from starlette.testclient import TestClient


__all__ = (
    "asgi_application",
    "HTMLResponse",
    "JSONResponse",
    "Path",
    "PathPrefix",
    "PlainTextResponse",
    "Response",
    "Router",
    "StreamingResponse",
    "Request",
    "TestClient",
)
__version__ = "0.1.6"
